/*
   filename: WCD.C

   Handy directory changer for MS-DOS and Unix.

   This program jumps to a (sub)directory.
   The directory to jump to can be given by only the first
   characters or with wildcards.

   wcd generates a treedata-file were it searches the directory.

   >wcd adir
   searches for a directory matching adir*
   wcd jumps to the last adir* matching directory found.
   A 100 % match of adir goes first and lets the user make a choise.

   Wildcards * and ? are supported.

Erwin Waterlander

Address : Philips Research Laboratories
	  Building WAY-4.13
	  Prof. Holstlaan 4
	  5656 AA Eindhoven, The Netherlands

E-mail  : waterlan@natlab.research.philips.com
 or     : waterlan@xs4all.nl
WWW     : http://www.xs4all.nl


======================================================================
= Copyright                                                          =
======================================================================
Copyright (C) 1997 Erwin Waterlander

This program is freeware.
Permission is granted to any individual or institution to use,
copy, redistribute or modify this program, provided this
copyright notice is retained.

For the DOS-16-bit and the Unix version I used DOSDIR. See copyright
below.
I used the matching algorithm of Info-Zip's unzip program. See copyright
below.

Both DOSDIR and UNZIP are distributed by the Simtel.Net(sm)
world-wide network

......................................................................

DOSDIR: A Portable DOS/UNIX/VMS
Directory Interface, version 2.1 by Jason Mathews,
NASA/Goddard Space Flight Center, Greenbelt, Maryland  20771-0001 USA

Copyright (C) 1994 Jason Mathews.

Permission is granted to any individual or institution to use, copy
or redistribute this software so long as it is not sold for profit,
provided this copyright notice is retained.
......................................................................

Copyright on recmatch() from Zip's match.c

Copyright (C) 1990-1992 Mark Adler, Richard B. Wales, Jean-loup Gailly,
Kai Uwe Rommel and Igor Mandrichenko.

Permission is granted to any individual or institution to use, copy,
or redistribute this software so long as all of the original files are
included unmodified, that it is not sold for profit, and that this copy-
right notice is retained.
=======================================================================
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "dosdir.h"
#include "match.h"
#include "dirent.h"

#ifdef VMS
#  define EXIT_OK 1
#  define ALT_SW || *argv[i]=='/'
#  define CHDIR(s) vms_chdir(s)
#else
#  define EXIT_OK 0
#  define CHDIR(s) chdir(s)
#  ifdef MSDOS
#    define ALT_SW || *argv[i]=='/'
#    define TREEFILE "\\treedata.wcd"
#    define ROOTDIR "\\"
#    define DIR_SEPERATOR '/'
#  else /* ?UNIX */
#    define ALT_SW
#    define TREEFILE "/.treedata.wcd"
#    define EXTRA_TREEFILE "/.extra.wcd"
#    define ROOTDIR "/"
#    define HOMESTRING "/home/"
#    define GO_FILE "/bin/wcd.go"
#    define DIR_SEPERATOR '/'
#  endif /* ?MSDOS */
#endif /* ?VMS */
#define MAX_LIST 22        /* max list of perfect match directories */
#define MAX_WILD_LIST 22   /* max list of wild match directories */
#define MAX_EF 10          /* max nr. of extra tree-files */
#define VERSION      "1.7e"
#define VERSION_DATE "Dec 9 1997"

typedef struct TDirTag {
	char* dirname;
	struct TDirTag *next;
} TDirEntry;

typedef struct {
	TDirEntry *head, *tail;
} TDirList;

/* Global variables */

FILE *outfile;
char filemask[DD_MAXFILE+DD_MAXEXT];
int  attrib = DD_NORMAL;
unsigned dir_files;
char dir[DD_MAXDIR];
char org_dir[DD_MAXDIR];
char best_match[DD_MAXDIR];
char pm[MAX_LIST][DD_MAXDIR];
char wm[MAX_WILD_LIST][DD_MAXDIR];
char no_match =1, perfect_match = 0;
int  perfect_match_nr = 0, wild_match_nr = 0;
#ifdef UNIX
 char quiet = 0;
#else
 char quiet = 1;
#endif
const char *default_mask = ALL_FILES_MASK;
int (*filematch) OF(( const char*, const char*, int ));

/* Function prototypes */

void finddirs    OF((char *dir));
void scanfile    OF((char *filename));
int  strcompare  OF((const char*, const char*, int));
void q_insert    OF((TDirList *list, const char *s));
int  q_remove    OF((TDirList *list, char *s));
void printhelp   OF((void));
int  wcd_get_int OF((void));
void empty_wcdgo OF ((char *go_file, char *rootdir));

#if defined (VMS)

int vms_chdir(char* dir)
{
	/* kill version number on directories */
	char* s = strchr(dir, ';');
	if (s) *s = '\0';
	return chdir(dir);
}

/*---------------------------------------------------------------------*
Name            strupr - converts a string to upper-case

Description     strupr converts lower-case letters in string str to upper-case.
		No other changes occur.
*---------------------------------------------------------------------*/
char* strupr(const char* s)
{
	register char *ps = s;
	while (*ps)
	{
		if (islower((unsigned char)*ps))
			*ps = _toupper((unsigned char)*ps);
		ps++;
	}
	return s;
}

#elif defined (MSDOS)

/* No valid MS-DOS directories start with a period (.),
 * except for .. or ., so anything with a period prefix
 * must be special.
 */
#  define SpecialDir(f) (*(f) == '.')

#else  /* ? UNIX */

int SpecialDir OF((const char *path));

/* Function: SpecialDir
 *
 * Purpose:  Test for special directories
 *
 * Returns: 1 if path = "." or ".."
 *          0 otherwise.
 */
int SpecialDir(path)
const char *path;
{
	if (*path != '.') return 0;
	if (*(++path) == '.') path++;
	return (*path=='/' || *path=='\0');
}

#endif /* ?VMS */

int strcompare(s1, s2, ignore_case)
const char *s1, *s2;
int ignore_case;
{
	return !strcmp(s1, s2);
}

/********************************************************************
 *
 *                    finddirs(dir)
 *
 ********************************************************************/

void finddirs(dir)
char* dir;
{
	static dd_ffblk fb;       /* file block structure */
	static char tmp[DD_MAXDIR];      /* tmp string buffer */
	int rc;                       /* error code */
	TDirList list;                /* directory queue */
	char *ptr;


	if (dir)
	{
#ifdef MSDOS
		int len = strlen(dir);
		/* strip ending separator (if present), which DOS doesn't like. */
		if (len > 1 && dir[len-1]==DIR_END) dir[len-1] = 0;
#endif /* ?MSDOS */
		if (CHDIR(dir)) return; /* ?err */
	}

	rc = dd_findfirst( default_mask, &fb, attrib | DD_DIREC );
	list.head = list.tail = 0;
	dir_files = 0;

	while (rc==0)
	{
		if (attrib & DD_DIREC && DD_ISDIREC(fb.dd_mode))
		{
#ifndef VMS
			/*  Ignore directory entries starting with '.'
	   *  which includes the current and parent directories.
	   */
			if (!SpecialDir(fb.dd_name))
#endif /* ?!VMS */
				q_insert(&list, fb.dd_name);
		}

		/* if match then do something with the file */
		if (filematch(fb.dd_name, filemask, 0))
		{
			if (!dir_files++)
			{
				int len;
				getcwd(tmp, sizeof(tmp));
#ifndef VMS
				len = strlen(tmp);
				if (len==0 || tmp[--len] == DIR_END)
					tmp[len] = '\0';
#endif /* ?!VMS */
#ifdef MSDOS
		/* Change backslashes to DIR_SEPERATOR
		   Avoid doing compares on strings
		   with backslashes. */
		    while ((ptr=strchr(tmp,'\\'))) * ptr=DIR_SEPERATOR;
			if ( (ptr=strstr(tmp,"/")) != NULL)
			fprintf(outfile,"%s\n",ptr);
#else
			fprintf(outfile,"%s\n", tmp);
#endif
			}
		}

		rc = dd_findnext(&fb);
	} /* while !rc */

	/* recursively parse subdirectories (if any) */
	while (q_remove(&list, tmp))
		finddirs(tmp);

	if (dir) chdir(DIR_PARENT); /* go to parent directory */
}

/******************************************************************
 *
 *          q_insert - insert directory name to queue
 *
 ******************************************************************/

void q_insert(list, s)
TDirList *list;
const char *s;
{
	TDirEntry *ptr;
	int len = strlen(s);
	if (!len) return;
	if ((ptr = (TDirEntry*) malloc(sizeof(TDirEntry))) == NULL )
	{
		perror("malloc");
		return;
	}
	if ((ptr->dirname = (char*) malloc(len+1)) == NULL )
	{
		perror("malloc");
		free(ptr);
		return;
	}
	strcpy(ptr->dirname, s);
	ptr->next = NULL;
	if (!list->head) list->head = ptr;
	else list->tail->next = ptr;
	list->tail = ptr;
}

/*******************************************************************
 *
 *         q_remove - remove directory name from queue
 *
 *******************************************************************/

int q_remove(list, s)
TDirList *list;
char *s;
{
	TDirEntry *ptr = list->head;
	if (!ptr) return 0;		/* queue empty? */
	strcpy(s, ptr->dirname);
	list->head = ptr->next;
	free(ptr->dirname);
	free(ptr);
	return 1;			/* okay */
}

/********************************************************************
 *
 *                    scanfile(char *filename)
 *
 *
 ********************************************************************/

void scanfile(filename)
char *filename;
{
	FILE *infile;
	char line[DD_MAXDIR];
	char help1_str[DD_MAXDIR];
	char c;
	char *line_end,*help_str;
	int  i,j;
	char error = 0;
#ifdef UNIX
	char *ptr;
#endif

	/* open treedata-file */
	if  ((infile = fopen(filename,"r")) == NULL)
	{
		fprintf(stderr,"Error opening treedata-file %s\n",filename);
	}
	else
	{

		if( (help_str = strrchr(org_dir,DIR_SEPERATOR)) != NULL)
		 help_str++;
		else help_str = org_dir;

		if (!dd_iswild(help_str))
			{
			  strcpy(dir,org_dir);
			  strcat(dir,"*");
			}

		while (!feof(infile) )
		{
			j=0;
			error = 0;
			for(i=0; ((c=getc(infile)) != '\n') && (c!=EOF) ;i++)
				if (i < (DD_MAXDIR-1))
				{
					line[i]=c;
					j++;
				}
				else error = 1;
			line[j]='\0';  /* end string */


			if(error)
				printf("Error: line too long ( %d > %d). Fix: increase DD_MAXDIR.\n",i,(DD_MAXDIR-1));


			if( (line_end = strrchr(line,DIR_SEPERATOR)) != NULL)
				line_end++;
			else
				line_end = line;

			/* does the search string have a DIR_SEPERATOR? */

			  if( (help_str = strrchr(org_dir,DIR_SEPERATOR)) != NULL)
			    help_str++;
			  else help_str = org_dir;


			  strcpy(help1_str,"*");
			  strcat(help1_str,dir);

		    /* test for a perfect match */
#ifdef MSDOS
			/* ignore case on a DOS platform */
			if ( (stricmp(line_end,help_str)==0) && (dd_match(line,help1_str,1)) )
#else
			if ( (strcmp(line_end,help_str)==0) && (dd_match(line,help1_str,0)) )
#endif
			{
#ifdef UNIX
	/* skip the /tmp_mnt string */
	if( (ptr=strstr(line,getenv("HOME")))!= NULL)
		strcpy(line,ptr);
#endif

				perfect_match =1;
				if(!quiet) printf("%s\n",line);
				if (perfect_match_nr<MAX_LIST) strcpy(pm[perfect_match_nr],line);
				strcpy(best_match,line);
				perfect_match_nr++;
				no_match = 0;
			}

			  if( (help_str = strrchr(dir,DIR_SEPERATOR)) != NULL)
			    help_str++;
			  else help_str = dir;


			/* test for a wild match if no perfect match */
#ifdef MSDOS
			/* ignore case on a DOS platform */
			if ( (dd_match(line_end,help_str,1)) && (dd_match(line,help1_str,1)) && !perfect_match)
#else
			if ( (dd_match(line_end,help_str,0)) && (dd_match(line,help1_str,0)) && !perfect_match)
#endif
			{
#ifdef UNIX
	/* skip the /tmp_mnt string */
	if( (ptr=strstr(line,getenv("HOME")))!= NULL)
		strcpy(line,ptr);
#endif
				if(!quiet) printf("%s\n",line);
				if (wild_match_nr<MAX_WILD_LIST) strcpy(wm[wild_match_nr],line);
				strcpy(best_match,line);
				wild_match_nr++;
				no_match = 0;
			}

		}   /* while (!feof(infile) ) */
		fclose(infile);
	}
}
/********************************************************************
 *
 *                 Get int
 *
 ********************************************************************/

int wcd_get_int()
{
  int i;
  char string[256];

    gets(string);
    fflush(stdin); /* flush the input stream in case of bad input */

    i=atoi(string);

  return(i);
}


/********************************************************************
 *
 *                 Print help
 *
 ********************************************************************/
void printhelp()
{
	printf("WCD  - Waterlander Change Directory %s\n",VERSION);
	printf("     - by Erwin Waterlander, %s\n\n",VERSION_DATE);
	printf("Usage:  wcd [dir] [-h] [-s] [-a] [-e] [-q] [-Q] [-c] [-v]\n");
	printf("            [-u <username>] [-f <treefile>]\n");
	printf("  dir (partial) name of directory to change to.\n");
	printf("      Wildcards *, ? and [SET] are supported!\n");
	printf("  -a  Add current directory to treedata\n");
	printf("  -e  add current directory to Extra treedata (Unix only)\n");
	printf("  -q  Quiet operation (Unix)     unQuiet operation (DOS)\n");
	printf("  -u  add treefile of other User (Unix only)\n");
	printf("  -f  add extra treeFile\n");
	printf("  -s  (re)Scan disk\n");
	printf("  -h  show this Help           -v  print Version info\n");
	printf("  -Q  Quieter operation        -c  direct CD mode\n");
	printf("\nExamples:\n");
	printf("   wcd Un -s               wcd src/wcd\n");
	printf("   wcd *ix                 wcd doe*/vhdl -u doe\n");
	printf("   wcd w*[1-3]             wcd src -q -f /home/doe/.extra.wcd\n\n");
	printf("e-mail: waterlan@natlab.research.philips.com\n");

#ifdef UNIX
	printf("---\nKorn Shell installation:\n");
	printf("Copy wcd.exe to your %s/bin directory.\n", getenv("HOME"));
	printf("Add the following function to your %s/.kshrc file.\n", getenv("HOME"));
	printf("   function wcd\n   {\n   $HOME/bin/wcd.exe $*\n   . $HOME/bin/wcd.go\n   }\n");
	printf("Start a new Korn Shell and go.\n");
	printf("---\nC Shell installation:\n");
	printf("Copy wcd.exe to your %s/bin directory.\n", getenv("HOME"));
	printf("Add the following alias to your %s/.cshrc file.\n", getenv("HOME"));
	printf("   alias wcd \"$HOME/bin/wcd.exe \\!* ; source $HOME/bin/wcd.go\"\n");
	printf("Start a new C Shell and go.\n");
	printf("---\nBourne Again Shell installation:\n");
	printf("Copy wcd.exe to your %s/bin directory.\n", getenv("HOME"));
	printf("Add the following function to your %s/.bashrc file.\n", getenv("HOME"));
	printf("   function wcd\n   {\n   $HOME/bin/wcd.exe $*\n   source $HOME/bin/wcd.go\n   }\n");
	printf("Start a new bash and go.\n");
	printf("---\nA default extra directory-list can be added in file %s%s\n", getenv("HOME"),EXTRA_TREEFILE);
	printf("Do not end the directory-names with a slash.\n");
#endif
}

/********************************************************************
 *
 *             empty wcd.go file
 *
 ********************************************************************/

#ifdef UNIX
void empty_wcdgo(go_file, help1_str)
char *go_file, *help1_str;
{
	if  ((outfile = fopen(go_file,"w")) == NULL)
	{
		fprintf(stderr,"Error opening file %s for write.\n", go_file);
		exit(0);
	}

	fprintf(outfile,"\n");
	fclose(outfile);

}
#endif
/********************************************************************
 *
 *                             MAIN
 *
 ********************************************************************/

int main(argc, argv)
int argc;
char** argv;
{
	int  i;
	int  flags = 0;
	char *path = 0;
	char rootdir[DD_MAXDIR],treefile[DD_MAXDIR];
	FILE *infile;
	char scan_disk = 0;
	char *ptr;
	char extra_files[MAX_EF][DD_MAXDIR];
	char nr_extra_files = 0;
	int  quieter = 0, cd = 0;

	int len;
	static char tmp[DD_MAXDIR];      /* tmp string buffer */
	DIR* dirp; /* GJM */


#ifdef MSDOS
	strcpy(rootdir,ROOTDIR);
	strcpy(treefile,TREEFILE);
#else
	int  j,k;
	char help1_str[DD_MAXDIR];
	char go_file[DD_MAXDIR];

	strcpy(rootdir,getenv("HOME"));
	strcat(rootdir,ROOTDIR);
	strcpy(treefile,getenv("HOME"));
	strcat(treefile,TREEFILE);
	strcpy(go_file,getenv("HOME"));
	strcat(go_file,GO_FILE);
#endif

	if (argc == 1 )
	{
		printhelp();

#ifdef UNIX       /* empty wcd.go file */
		empty_wcdgo(go_file,help1_str);
#endif

		return EXIT_OK;
	}

	strcpy(dir,"");
	strcpy(org_dir,"");

	attrib |= DD_DIREC;  /* set recursive search on */

	/* ---------------------- parse the commandline ------------*/

	for (i=1; i < argc; i++)
	{
		if (*argv[i]=='-' ALT_SW) /* is it a switch? */
			switch (argv[i][1]) {
			case 'S':
			case 's':
				  scan_disk = 1;
				break;
			case 'A':
			case 'a':
				getcwd(tmp, sizeof(tmp));
				if(tmp != NULL)
				{
				  len = strlen(tmp);
				  if (len==0 || tmp[--len] == DIR_END)
					tmp[len] = '\0';

				  /* open the treedata file */
				  if  ((outfile = fopen(treefile,"a")) == NULL)
				  {
				   fprintf(stderr,"Error opening file %s for write.\n", treefile);
				   exit(0);
				  }
				  else
				  {
				   while ((ptr=strchr(tmp,'\\'))) * ptr= DIR_SEPERATOR;
				   if ( (ptr=strstr(tmp,"/")) != NULL)
				   {
				   	fprintf(outfile,"%s\n",ptr);
				    printf("%s added.\n",ptr);
				   }
				  }
				  fclose(outfile);
				}
				break;
#ifdef UNIX
			case 'E':
			case 'e':
				getcwd(tmp, sizeof(tmp));
				if(tmp != NULL)
				{
				  len = strlen(tmp);
				  if (len==0 || tmp[--len] == DIR_END)
					tmp[len] = '\0';

				strcpy(help1_str,getenv("HOME"));
				strcat(help1_str,EXTRA_TREEFILE);
				  /* open the treedata file */
				  if  ((outfile = fopen(help1_str,"a")) == NULL)
				  {
				   fprintf(stderr,"Error opening file %s for write.\n", help1_str);
				   exit(0);
				  }
				  else
				  {
				   while ((ptr=strchr(tmp,'\\'))) * ptr= DIR_SEPERATOR;
				   fprintf(outfile,"%s\n",tmp);
				   printf("%s added.\n",tmp);
				  }
				  fclose(outfile);
				}
				break;
#endif
			case 'q':
#ifdef UNIX
				  quiet = 1;
#else
				  quiet = 0;
#endif
				break;
			case 'Q':
				  quiet = 1;
				  quieter = 1;
				break;
			case 'v':
			case 'V':
				printf("WCD %s, %s\n\n",VERSION,VERSION_DATE);
				printf("WCD is freeware.\n\n");
				printf("Copyright (C) 1997 Erwin Waterlander\n");
				printf("Copyright (C) 1994 Jason Mathews on DOSDIR\n");
				printf("Copyright (C) 1990-1992 Mark Adler, Richard B. Wales, Jean-loup Gailly,\n");
				printf("Kai Uwe Rommel and Igor Mandrichenko on recmatch()\n\n");
				printf("See also file wcd.txt\n");
				break;
			case 'c':
			case 'C':
				  cd = 1;
				break;
			case 'f':
				break;
#ifdef UNIX
			case 'u':
				break;
#endif
			default:               /* any switch except the above */
				printhelp();

#ifdef UNIX       /* empty wcd.go file */
				empty_wcdgo(go_file,help1_str);
#endif

				return EXIT_OK;
			}
		else /* Not a switch. Must be a dir of filename. */
		{


		    if ((strcmp(argv[i-1],"-f") == 0 )
		       && (nr_extra_files < MAX_EF))
			{
			  strcpy(extra_files[nr_extra_files],argv[i]);
			  nr_extra_files++;
			 }
			 else
#ifdef UNIX
		    if ((strcmp(argv[i-1],"-u") == 0 )
		       && (nr_extra_files < MAX_EF))
			{
			  strcpy(extra_files[nr_extra_files],HOMESTRING);
			  strcat(extra_files[nr_extra_files],argv[i]);
			  strcat(extra_files[nr_extra_files],TREEFILE);
			  nr_extra_files++;
			 }
			 else
#endif
			 {
				      strcpy(dir,argv[i]);
				      strcpy(org_dir,argv[i]);
#ifdef MSDOS
		/* Change backslashes to DIR_SEPERATOR if a backslash was typed
		   Avoid doing compares on strings
		   with backslashes. */
				  while ((ptr=strchr(dir,'\\'))) * ptr= DIR_SEPERATOR;
				  while ((ptr=strchr(org_dir,'\\'))) *ptr= DIR_SEPERATOR;
#endif
				      flags = DIRECTORY;
				      }

		}


	} /* for */


	/*--- end parsing commandline -----------*/



	if ((cd==1)&&(strcmp(dir,"")!=0)) /* Try open dir direct. */
	{
		if((dirp=opendir(org_dir)) != NULL) /* GJM */
		{
			closedir(dirp); /* GJM */
			strcpy(best_match,org_dir);
			if (!quieter)
	   			printf("-> %s\n",best_match);
#ifdef UNIX
	j = strlen(best_match);
	k = 0;
	for (i=0; i < j ; i++)
	{ if ( (best_match[i] == '$') || (best_match[i] == '"') )
	   {
	   help1_str[k] = '\\'; k++;
	   }
	  help1_str[k] = best_match[i];
	  k++ ;
	}
	help1_str[k] = '\0' ;

	strcpy(best_match,"\"");
	strcat(best_match,help1_str);
	strcat(best_match,"\"");



		if  ((outfile = fopen(go_file,"w")) == NULL)
		{
			fprintf(stderr,"Error opening file %s for write.\n", go_file);
			exit(0);
		}
		fprintf(outfile,"cd %s\n", best_match);
		fclose(outfile);

#else
		CHDIR(best_match); /* change to directory */
#endif
		return EXIT_OK;
		}



	}


	flags |= WILDCARDS;
	strcpy(filemask, default_mask);

#if defined (VMS)
	if (!strchr(filemask, ';')) strcat(filemask, ";*");
#endif /* ?VMS */

	/* set comparison function */
	filematch = (flags & WILDCARDS) ? *dd_match : *strcompare;


	/* does treedata-file exist? */
	if  ((infile = fopen(treefile,"r")) == NULL)
		scan_disk = 1;
	else fclose(infile);


	if (scan_disk)     /* (re)Scan the disk? */
	{
		printf("Please wait. (re)Scanning disk. Building treedata-file.\n");

#ifdef MSDOS
	/* open the output file */
	if  ((outfile = fopen(treefile,"w")) == NULL)
	  if  ( (ptr = getenv("TEMP")) != NULL )
	   { strcpy(treefile,ptr);
	     strcat(treefile,TREEFILE);
	     if ((outfile = fopen(treefile,"w")) == NULL)
		if  ( (ptr = getenv("TMP")) != NULL )
		 { strcpy(treefile,ptr);
		   strcat(treefile,TREEFILE);
		   if ((outfile = fopen(treefile,"w")) == NULL)
		     if  ( (ptr = getenv("TMPDIR")) != NULL )
		       { strcpy(treefile,ptr);
				     strcat(treefile,TREEFILE);
			  if ((outfile = fopen(treefile,"w")) == NULL)
				   {
					 fprintf(stderr,"Error opening treefile for write.\n");
					 exit(0);
				   }
		       }
		  }
	   }
#else
		/* open the output file */
		if  ((outfile = fopen(treefile,"w")) == NULL)
		{
			fprintf(stderr,"Error opening file %s for write.\n", treefile);
			exit(0);
		}
#endif

		path=rootdir;
		finddirs( path ); /* Build treedata-file */
		fclose(outfile);

	}

	if (strcmp(dir,"")==0) /* Is there a directory to go to? */
	{
#ifdef UNIX       /* empty wcd.go file */
		empty_wcdgo(go_file,help1_str);
#endif
		return EXIT_OK;
	}


  scanfile(treefile); /* scan the treedata file */


#ifdef UNIX

  strcpy(treefile,getenv("HOME"));
  strcat(treefile,EXTRA_TREEFILE);
  if  ((outfile = fopen(treefile,"r")) != NULL)
	{
	   fclose(outfile);
	   scanfile(treefile); /* scan the extra treedata file */
	}

#endif

  /* search extra files */

  for (i=0;i<nr_extra_files;i++)
  {
   scanfile(extra_files[i]); /* scan the extra treedata file */
  }


	/* search is done */


	if ((perfect_match_nr > 1) &&(perfect_match_nr <= MAX_LIST))
	{
		for (i=0;(i<perfect_match_nr)&&(i<MAX_LIST);i++)
			printf("%d  %s\n",i+1,pm[i]);
		i= -1;
		printf("\nPerfect match for %d directories.\n",perfect_match_nr);

		printf("Please choose one (<Enter> to abort): ");
		i = wcd_get_int();

		if ( (i>0) && (i<=perfect_match_nr))
		{ i--;
		  strcpy(best_match,pm[i]);
		}
		else
		{
#ifdef UNIX       /* empty wcd.go file */
		empty_wcdgo(go_file,help1_str);
#endif
		return EXIT_OK;
		}

	}
	else if ((perfect_match==0)&&(wild_match_nr > 1)&&(wild_match_nr <= MAX_WILD_LIST))
	{
		for (i=0;(i<wild_match_nr)&&(i<MAX_WILD_LIST);i++)
			printf("%d  %s\n",i+1,wm[i]);
		i= -1;
		printf("\nWild match for %d directories.\n",wild_match_nr);

		printf("Please choose one (<Enter> to abort): ");
		i=wcd_get_int();

		if ( (i>0) && (i<=wild_match_nr))
		{
		  i--;
		  strcpy(best_match,wm[i]);
		}
		else
		{
#ifdef UNIX       /* empty wcd.go file */
		empty_wcdgo(go_file,help1_str);
#endif
		return EXIT_OK;
		}

	} else if ((perfect_match==0)&&(wild_match_nr > MAX_WILD_LIST))
		no_match = 1;


	/* #ifdef UNIX */
	/* skip the /tmp_mnt string */


#ifdef UNIX
	if( (ptr=strstr(best_match,HOMESTRING))!= NULL)
		strcpy(best_match,ptr);
#endif


	/*******************************/


	if ((no_match)&&(wild_match_nr > MAX_WILD_LIST))
	{  /* No perfect match, to many wild matches */

		printf("More than %d directories found matching %s\n",MAX_WILD_LIST,dir);
		printf("Try a more unique searchstring.\n");

#ifdef UNIX       /* empty wcd.go file */
		empty_wcdgo(go_file,help1_str);
#endif

	}
	else if ((no_match)&&(cd==0)) /* No match at all */
	{

		if((dirp=opendir(org_dir)) == NULL) /* GJM */
		{
			printf("No directory found matching %s\nPerhaps you need to rescan the disk.\n",dir);
#ifdef UNIX       /* empty wcd.go file */
			empty_wcdgo(go_file,help1_str);
#endif
		}
		else
		{
			closedir(dirp); /* GJM */
			no_match = 0;
			strcpy(best_match,org_dir);
		}


	}else if (no_match)
	  	printf("No directory found matching %s\nPerhaps you need to rescan the disk.\n",dir);


	/*******************************/


	if ((no_match==0)&&(perfect_match_nr > MAX_LIST))
	{  /* To many perfect matches */

		printf("More than %d directories found matching %s\n",MAX_LIST,org_dir);
		printf("Use subdirectory definition. Eg. wcd src/wcd\n");

#ifdef UNIX       /* empty wcd.go file */
		empty_wcdgo(go_file,help1_str);
#endif
	}
	else
	if (no_match==0) /* yes, a match */
	{
	if (!quieter)
	printf("-> %s\n",best_match);

#ifdef UNIX
	j = strlen(best_match);
	k = 0;
	for (i=0; i < j ; i++)
	{ if ( (best_match[i] == '$') || (best_match[i] == '"') )
	   {
	   help1_str[k] = '\\'; k++;
	   }
	  help1_str[k] = best_match[i];
	  k++ ;
	}
	help1_str[k] = '\0' ;

	strcpy(best_match,"\"");
	strcat(best_match,help1_str);
	strcat(best_match,"\"");



		if  ((outfile = fopen(go_file,"w")) == NULL)
		{
			fprintf(stderr,"Error opening file %s for write.\n", go_file);
			exit(0);
		}
		fprintf(outfile,"cd %s\n", best_match);
		fclose(outfile);

#else
		CHDIR(best_match); /* change to directory */
#endif
	}
	return EXIT_OK;
}
